"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.analyze = void 0;
const fs_1 = __importDefault(require("fs"));
const globby_1 = __importDefault(require("globby"));
const parse_1 = require("./parse");
const sleep_1 = require("./sleep");
/**
 * Return a Promise with the JSX usage analysis of a project
 *
 * This function can easily return several megabytes of data. Parsing every
 * single JS/TS file in a project can take a while. Please be patient. The
 * `onFile` callbacks are async. Parsing each file is not an async action and
 * may block for a couple seconds if the file is large. You can insert a "sleep"
 * command into the `onFile` callback if you would like to return control
 * briefly.
 */
function analyze({ babelPlugins = [], components, directory, files = ["**/*.{js,jsx,tsx}"], gitignore = true, ignore = [], onFile, prop, }) {
    return __awaiter(this, void 0, void 0, function* () {
        const timeStart = Date.now();
        const filenames = yield globby_1.default(files || "**/*.{js,jsx,tsx}", {
            absolute: true,
            onlyFiles: true,
            gitignore,
            ignore,
            cwd: directory || process.cwd(),
        });
        const reporter = new Reporter();
        for (const filename of filenames) {
            if (onFile) {
                yield onFile(filename);
            }
            else {
                yield sleep_1.sleep();
            }
            try {
                const code = fs_1.default.readFileSync(filename, "utf8");
                parse_1.parse(code, {
                    babelPlugins,
                    typescript: filename.endsWith(".tsx") || filename.endsWith(".ts"),
                    onlyComponents: components,
                    onComponent: (componentName) => {
                        reporter.addComponent(componentName);
                    },
                    onProp: ({ componentName, propName, propCode, startLoc, endLoc, propValue, }) => {
                        if (prop) {
                            let wantPropKey = prop;
                            let wantPropValue = undefined;
                            if (prop.includes("=")) {
                                const index = prop.indexOf("=");
                                const key = prop.slice(0, index);
                                const val = prop.slice(index + 1);
                                wantPropKey = key;
                                wantPropValue = val;
                            }
                            if (propName !== wantPropKey) {
                                return;
                            }
                            if (wantPropValue !== undefined && propValue !== wantPropValue) {
                                return;
                            }
                        }
                        const prettyCode = formatPrettyCode(code, startLoc.line, endLoc.line);
                        reporter.addProp(componentName, propName, {
                            propCode,
                            startLoc,
                            endLoc,
                            prettyCode,
                            filename,
                        });
                    },
                });
            }
            catch (error) {
                if (error instanceof SyntaxError) {
                    reporter.addParseError(filename, error);
                }
                else {
                    throw error;
                }
            }
        }
        const elapsedTime = (Date.now() - timeStart) / 1000;
        return {
            filenames,
            componentTotal: reporter.getComponentTotal(),
            componentUsageTotal: reporter.getComponentUsageTotal(),
            componentUsage: reporter.components,
            propUsage: reporter.props,
            lineUsage: reporter.lines,
            errors: reporter.errors,
            suggestedPlugins: reporter.suggestedPlugins,
            elapsedTime: elapsedTime,
        };
    });
}
exports.analyze = analyze;
////////////////////////////////////////////////////////////////////////////////
//
// Implementation details below
//
////////////////////////////////////////////////////////////////////////////////
const linesCache = new Map();
function getLines(code) {
    let lines = linesCache.get(code);
    if (lines) {
        return lines;
    }
    lines = code.split(/\r?\n/);
    linesCache.set(code, lines);
    return lines;
}
function formatPrettyCode(code, startLine, endLine) {
    const output = [];
    const lines = getLines(code);
    // Line numbers should be padded to at least 4 digits for consistency and
    // readability, but let's also let them grow if we have super long files :|
    const maxDigits = Math.max(String(startLine).length, String(endLine).length, 4);
    for (let lineno = startLine; lineno <= endLine; lineno++) {
        output.push(String(lineno).padStart(maxDigits) + " | " + lines[lineno - 1]);
    }
    return output.join("\n");
}
class Reporter {
    constructor() {
        this.components = {};
        this.props = {};
        this.lines = {};
        this.suggestedPlugins = [];
        this.errors = {};
    }
    // TODO: Does Babel expose the actual error type for us?
    addParseError(filename, error) {
        this.errors[filename] = {
            message: error.message,
            pos: error.pos,
            loc: error.loc,
            missingPlugin: error.missingPlugin,
        };
        for (const plugin of error.missingPlugin || []) {
            if (!this.suggestedPlugins.includes(plugin)) {
                this.suggestedPlugins.push(plugin);
            }
        }
    }
    addComponent(componentName) {
        this.components[componentName] = (this.components[componentName] || 0) + 1;
    }
    _incrementProp(componentName, propName) {
        if (!this.props[componentName]) {
            this.props[componentName] = {};
        }
        const count = this.props[componentName][propName] || 0;
        this.props[componentName][propName] = count + 1;
    }
    _ensureLines(componentName, propName) {
        if (!this.lines[componentName]) {
            this.lines[componentName] = {};
        }
        if (!this.lines[componentName][propName]) {
            this.lines[componentName][propName] = [];
        }
    }
    addProp(componentName, propName, line) {
        this._incrementProp(componentName, propName);
        this._ensureLines(componentName, propName);
        this.lines[componentName][propName].push(line);
    }
    getComponentTotal() {
        return Object.keys(this.components).length;
    }
    getComponentUsageTotal() {
        return Object.values(this.components).reduce((a, b) => a + b, 0);
    }
}
