"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.parse = void 0;
const parser_1 = require("@babel/parser");
const traverse_1 = __importDefault(require("@babel/traverse"));
const EXPRESSION = Symbol("formatPropValue.EXPRESSION");
function formatPropValue(value) {
    if (value === null) {
        return "true";
    }
    if (!value) {
        return EXPRESSION;
    }
    switch (value.type) {
        // TODO: Should we interpret anything else here?
        case "StringLiteral":
            return value.value;
        case "JSXExpressionContainer":
            return formatPropValue(value.expression);
        case "NumericLiteral":
        case "BooleanLiteral":
            return String(value.value);
        default:
            return EXPRESSION;
    }
}
function getAttributeName(attributeNode) {
    switch (attributeNode.type) {
        case "JSXIdentifier":
            return attributeNode.name;
        case "JSXNamespacedName":
            return attributeNode.name.name;
        case "JSXAttribute":
            return getAttributeName(attributeNode.name);
        case "JSXSpreadAttribute":
            return "{...}";
        default:
            throw new Error(`unexpected node type: ${attributeNode.type}`);
    }
}
function createProp(attributeNode) {
    return getAttributeName(attributeNode);
}
function getDottedName(nameNode) {
    switch (nameNode.type) {
        case "JSXMemberExpression":
            return [nameNode.object, nameNode.property].map(getDottedName).join(".");
        case "JSXIdentifier":
            return nameNode.name;
        default:
            throw new Error(`unexpected node type: ${nameNode.type}`);
    }
}
function createComponent(componentNode) {
    return getDottedName(componentNode.openingElement.name);
}
function parse(code, options = {}) {
    const { typescript = false, babelPlugins = [], onlyComponents = [], onComponent = () => { }, onProp = () => { }, } = options;
    function doReportComponent(component) {
        if (onlyComponents.length === 0) {
            return true;
        }
        return onlyComponents.indexOf(component) !== -1;
    }
    const ast = parser_1.parse(code, {
        sourceType: "unambiguous",
        allowReturnOutsideFunction: true,
        plugins: [
            typescript ? "typescript" : "flow",
            "jsx",
            "dynamicImport",
            "classProperties",
            "objectRestSpread",
            ...babelPlugins,
        ],
    });
    traverse_1.default(ast, {
        JSXElement(path) {
            const node = path.node;
            const componentName = createComponent(node);
            if (doReportComponent(componentName)) {
                onComponent(componentName);
                for (const propNode of node.openingElement.attributes) {
                    const propCode = code.slice(propNode.start || 0, propNode.end || -1);
                    const propName = createProp(propNode);
                    if (!propNode.loc) {
                        throw new Error(`JSXElement propNode.loc is missing`);
                    }
                    const { start: startLoc, end: endLoc } = propNode.loc;
                    const propValue = propNode.type === "JSXAttribute"
                        ? formatPropValue(propNode.value)
                        : formatPropValue(null);
                    onProp({
                        componentName,
                        propName,
                        propCode,
                        startLoc,
                        endLoc,
                        propValue,
                    });
                }
            }
        },
    });
}
exports.parse = parse;
